import os
import unittest

from configurati.loaders.utils import *


class SubstituteTest(unittest.TestCase):

  def test_no_substitution(self):
    assert substitute("abc") == "abc"

  def test_substitution(self):
    assert substitute("`1`") == 1

  def test_substitution_with_semicolon(self):
    assert substitute("`import os; os.getcwd()`") == os.getcwd()

  def test_substitution_with_newlines(self):
    assert substitute("`import os\nos.getcwd()`") == os.getcwd()

  def test_substitution_function(self):
    self.assertEqual(substitute("`lambda x: x + 1`")(1), 2)



from __future__ import absolute_import

import json

from .utils import substitute
from ..utils import recursive_apply


def load(f):
  contents = json.load(f)
  return recursive_apply(contents, value_func=substitute)



from .json import load as load_json
from .yaml import load as load_yaml
from .python import load as load_py
from .commandline import load as load_cl

from ..exceptions import ConfiguratiException


__all__ = [
  'load'
]


def load(f):
  if isinstance(f, list):
    return load_cl(f)
  else:
    n = f.name.lower()
    if n.endswith(".py"):
      return load_py(f)
    elif n.endswith(".yaml"):
      return load_yaml(f)
    elif n.endswith(".json"):
      return load_json(f)
    else:
      raise ConfiguratiException("Unrecognized file type {}".format(n))



"""
Commands to load a configuration
"""
import codecs
import os
import sys

from .attrs import attrs
from .loaders import load
from .utils import normalize_keys, previous_frame, add_globals, strip_invalid_keys
from .validation import is_spec


def load_config(path, relative_to_caller=False):
  """Load a config module as a dict

  Parameters
  ----------
  path : str
      path to configuration file
  relative_to_caller : bool
      if path isn't an absolute file path, interpret it as relative to the file
      from which this method is called.
  """
  if relative_to_caller and not os.path.abspath(path) == path:
    previous_fname  = previous_frame().f_globals['__file__']
    previous_fname  = os.path.abspath(previous_fname)
    previous_folder = os.path.split(previous_fname)[0]
    path            = os.path.join(previous_folder, path)
  with codecs.open(path, 'r', 'UTF-8') as f:
    return attrs.from_dict(strip_invalid_keys(normalize_keys(load(f))))


def load_spec(path, relative_to_caller=False):
  """Load a configuration specification"""
  spec = load_config(path, relative_to_caller=relative_to_caller)
  spec = { k:v for k,v in spec.items()
           if is_spec(v) }
  return attrs.from_dict(spec)


def import_config(path, relative_to_caller=True):
  """Load contents of another configuration file into the environment

  Parameters
  ----------
  path : str
      path to configuration file
  relative_to_caller : bool
      if path isn't an absolute file path, interpret it as relative to the file
      from which this method is called.
  """
  variables = load_config(path, relative_to_caller=relative_to_caller)
  add_globals(**variables)


def import_spec(path, relative_to_caller=True):
  """Import another specification's contents into the environment"""
  variables = load_spec(path, relative_to_caller=relative_to_caller)
  add_globals(**variables)


def env(variable_name):
  """Load an environment variable

  Parameters
  ----------
  variable_name : str
      environment variable name
  """
  if variable_name in os.environ:
    return os.environ[variable_name]
  else:
    return None




"""
The `configure` command
"""
import sys

from .loaders import load
from .commands import load_config, load_spec
from .globals import CONFIG
from .utils import update
from .validation import validate


def configure(args=None, config=None, spec=None):
  # load command line arguments
  if args is None:
    args = sys.argv[1:]
  if args is not None:
    args = load(args)
  else:
    args = {}

  # load configuration file
  if config is None and 'config' in args:
    config = args['config']
  if config is not None:
    result = update(args, load_config(config))
  else:
    result = args

  # apply config spec
  if spec is None and 'spec' in args:
    spec = args['spec']
  if spec is not None:
    result = validate(load_spec(spec), result)

  # set globally and return
  CONFIG(result)
  return CONFIG()



import unittest

from configurati.loaders.commandline import *
from configurati.utils import Missing


class NextTests(unittest.TestCase):

  def test_space_separated(self):
    assert next(["--key", "value"]) == ("key", "value", [])

  def test_equals_separated(self):
    assert next(["--key=value"]) == ("key", "value", [])

  def test_json_value(self):
    assert next(["--key", '{"a": 1}']) == ("key", {'a': 1}, [])


class LoadTests(unittest.TestCase):

  def test_simple(self):
    assert load(["--a", "1"]) == {"a": 1}

  def test_multiple(self):
    assert load(["--a", "1", "--b=cde"]) == {"a": 1, "b": "cde"}

  def test_nested_object(self):
    assert load(["--a.b", "1"]) == {"a": {"b": 1}}

  def test_nested_object2(self):
    # XXX shouldn't this be {"a": {"b": [Missing, 1]}} ?
    assert load(["--a.b[1]", "1"]) == {"a": {"b": [Missing, 1]}}



from __future__ import absolute_import

import unittest
import yaml

from configurati.loaders.yaml import *
from configurati.loaders.tests import LoadTests


class YamlLoadTest(LoadTests, unittest.TestCase):

  def setUp(self):
    self.loadfunc  = load
    self.dumpfunc  = yaml.dump
    self.extension = "yaml"



from StringIO import StringIO
import os
import unittest


class LoadTests(object):
  """Test a load-from-file module"""

  def test_load(self):
    o = {
      'a': "`import os; os.getcwd()`",
      'b': {
        'c': [1,2,3],
        'd': "abc"
      }
    }

    # write to fake file
    s = StringIO()
    s.name = "test." + self.extension
    self.dumpfunc(o, s)
    s.seek(0)

    # load from file
    o2 = self.loadfunc(s)

    assert o2['a'] == os.getcwd()
    assert o2['b'] == o['b']



from __future__ import absolute_import

import json

from ..attrs import set
from ..exceptions import ConfiguratiException


def load(args):
  """Construct configuration object from command line arguments alone"""
  result = {}
  if args is None:
    args = sys.argv[1:]
  while len(args) > 0:
    k, v, args = next(args)
    set(result, "." + k, v, build=True)
  return result


def next(args):
  """Get next key-value pair from arguments"""
  a = args[0]
  if not a.startswith('--'):
    raise ConfiguratiException('Arguments must be of the form "--key value" or "--key=value"')

  a = a[2:]
  if '=' in a:
    # "--key=value"
    key, value = a.split("=", 1)
    remaining = args[1:]
  else:
    # "--key" "value"
    if not len(args) >= 2:
      raise ConfiguratiException('Found key but no value: "{}"'.format(args))
    key, value = a, args[1]
    remaining = args[2:]

  return key, parse_value(value), remaining


def parse_value(s):
  """Parse a value expression"""
  # XXX should this be any Python expression?
  try:
    s = json.loads(s)
  except ValueError:
    pass
  return s



import argparse
from StringIO import StringIO
import sys
import uuid

import json
import yaml

from .loaders import load


if __name__ == '__main__':
  parser = argparse.ArgumentParser(
      "evaluate configurati config file"
    )
  parser.add_argument("format", choices=["yaml", "json"],
      help="Format of config file")
  args = parser.parse_args()

  # write data to file-like object
  s = StringIO()
  s.write(sys.stdin.read())
  s.seek(0)
  s.name = str(uuid.uuid4()) + "." + args.format

  # load and evaluate config
  config = load(s)

  # write it back out

  if args.format == 'yaml':
    print yaml.dump(config)
  elif args.format == 'json':
    print json.dumps(config)



import itertools
import re
import sys


def identity(x):
  return x


def recursive_apply(obj, key_func=identity, value_func=identity):
  """Recursively apply a function to the keys and values of a nested object"""
  if isinstance(obj, dict):
    result = {}
    for k, v in obj.items():
      k = key_func(k)
      result[k] = recursive_apply(v, key_func, value_func)
    return value_func(result)
  elif isinstance(obj, list):
    result = []
    for v in obj:
      result.append(recursive_apply(v, key_func, value_func))
    return result
  elif isinstance(obj, tuple):
    result = []
    for v in obj:
      result.append(recursive_apply(v, key_func, value_func))
    return tuple(result)
  else:
    return value_func(obj)


def normalize_keys(obj):
  """replace list-variable with list_variable"""

  def key_func(k):
    return re.sub("([a-zA-Z])-([a-zA-Z])", r"\1_\2", k)

  return recursive_apply(obj, key_func=key_func)


def strip_invalid_keys(obj):
  # only keys of this format are valid
  valid_identifier = re.compile("""^[a-zA-Z][a-zA-Z0-9_-]*$""")

  if isinstance(obj, dict):
    result = {}
    for k, v in obj.items():
      if valid_identifier.search(k) is not None:
        result[k] = strip_invalid_keys(v)
    return result
  elif isinstance(obj, list):
    return [ strip_invalid_keys(v) for v in obj ]
  elif isinstance(obj, tuple):
    return tuple( strip_invalid_keys(v) for v in obj  )
  else:
    return obj


def update(o1, o2):
  """Overlay `o1` over `o2`"""

  def update_dict(o1, o2):
    if not isinstance(o2, dict):
      o2 = {}
    for k, v in o1.items():
      o2[k] = update(o1[k], o2.get(k, {}))
    return o2

  def update_list(o1, o2):
    if not isinstance(o2, list):
      try:
        o2 = list(o2)
      except ValueError:
        o2 = []
    if len(o1) > len(o2):
      o2 = o2 + [Missing] * (len(o1) - len(o2))
    if len(o2) > len(o1):
      o1 = o1 + [Missing] * (len(o2) - len(o1))
    for i in range(len(o1)):
      o2[i] = update(o1[i], o2[i])
    return o2

  def update_tuple(o1, o2):
    return tuple(update_list(o1, list(o2)))


  if isinstance(o1, dict):
    return update_dict(o1, o2)
  elif isinstance(o1, list):
    return update_list(o1, o2)
  elif isinstance(o1, tuple):
    return update_tuple(o1, o2)
  else:
    if isinstance(o1, Missing_):
      return o2
    else:
      return o1


class Missing_(object):
  """Represents an unspecified value in a list or tuple"""

  def __str__(self):
    return "Missing"

  def __repr__(self):
    return str(self)


Missing = Missing_()

def previous_frame():
  """Find the first frame in the call stack not originating from this module"""
  module_name = __name__.split(".")
  for i in itertools.count():
    frame = sys._getframe(i)
    frame_name = frame.f_globals['__name__'].split(".")
    if frame_name[0] != module_name[0]:
      # first frame outside of this library
      return frame
    else:
      # XXX configurati tests need this to work
      if 'tests' in frame_name:
        return frame


def add_globals(**kwargs):
  """Update global dictionary of whoever is using this module"""
  # get first stack frame not in this module
  # update environment of said frame
  glob = previous_frame().f_globals
  glob.update(kwargs)



import os
from tempfile import NamedTemporaryFile as NTF
import unittest

from configurati.loaders.python import load


class PythonLoadTests(unittest.TestCase):

  def test_load(self):
    s = "\n".join([
      "import os",
      "a = os.getcwd()",
      "b = {",
      "   'c': [1,2,3],",
      "   'd': 'abc'",
      "  }",
    ])

    with NTF(suffix='.py') as f:
      # write to file
      f.write(s)
      f.flush()
      f.seek(0)

      # load it
      o2 = load(f)

      assert o2['a'] == os.getcwd()
      assert o2['b'] == {"c": [1,2,3], "d": "abc"}



import imp
import os
import uuid


def load(f):
  modname = str(uuid.uuid4())
  module = imp.load_source(modname, os.path.abspath(f.name))
  variables = {
    k:getattr(module, k) for k in dir(module)
    if not '__' in k
  }

  return variables



from .attrs import *
from .configure import *
from .commands import *
from .exceptions import *
from .globals import *
from .validation import *
from .utils import *


__version__ = "0.2.3"

__all__ = [
    # attrs
    'attrs',

    # commands
    'env',
    'import_config',
    'import_spec',
    'load_config',
    'load_spec',

    # configure
    'configure',

    # exceptions
    'ConfiguratiException',
    'ValidationError',

    # globals
    'CONFIG',

    # validation
    'one_of',
    'optional',
    'required',

    # utils
    'Missing'
]



from __future__ import absolute_import

import json
import unittest

from configurati.loaders.json import *
from configurati.loaders.tests import LoadTests


class JsonLoadTests(LoadTests, unittest.TestCase):

  def setUp(self):
    self.loadfunc  = load
    self.dumpfunc  = json.dump
    self.extension = "json"



import code
import re


def substitute(s):
  """Contents of `...` evaluated in Python"""
  if isinstance(s, basestring) and s.count("`") == 2:
    match = re.search("""^`([^`]+)`$""", s)
    contents, rest = match.group(1), s[match.end():]
    return evaluate(contents)
  else:
    return s


def evaluate(line):
  """Evaluate a line and return its final output"""
  # XXX this isn't smart enough to know about semicolons/newlines in strings,
  # or code where the final result relies on indentation
  line = re.split(";|\n", line)
  line[-1] = "OUTPUT = " + line[-1]
  line = ";".join(line)

  interpreter = code.InteractiveConsole()
  interpreter.push(line)
  return interpreter.locals["OUTPUT"]



from __future__ import absolute_import

try:
  import yaml
except ImportError:
  import sys
  sys.stderr.write("Unable to import `yaml`. Install with `pip install PyYAML`\n")
  sys.exit(1)

from .utils import substitute
from ..utils import recursive_apply


def load(f):
  contents = yaml.load(f)
  return recursive_apply(contents, value_func=substitute)



"""
Global way to access configuration.
"""

from .attrs import attrs


# globally accessible configuration object.
# accessible via CONFIG()
_CONFIG = attrs()


def CONFIG(config=None):
  if config is not None:
    global _CONFIG
    _CONFIG = attrs.from_dict(config)
  return _CONFIG



"""
Tools for validating a configuration spec
"""

from .attrs import attrs
from .exceptions import ValidationError
from .utils import identity, Missing


class variable(object):
  def __init__(self, type=str, help=None):
    self.type = type
    self.help = help

  def __repr__(self):
    return str(self)

  def __eq__(self, other):
    return isinstance(other, variable) and \
        other.type == self.type and \
        other.help == self.help


class optional(variable):
  """An optional variable

  Used in a spec file to denote a variable that has a default value and may be
  left unspecified in a config file.

  >>> version = optional(default="1.0-SNAPSHOT", \
                         type=str, \
                         help="Version of this application")

  Parameters
  ----------
  type : function
      function to coerce argument to desired type
  help : str
      long description of this variable
  default : object
      default value to use if this variable isn't in the config.
  """
  def __init__(self, type=identity, help=None, default=None):
    super(optional, self).__init__(type, help)
    self.default = default

  def __str__(self):
    return (
        "optional(type=%s, help=%s, default=%s)" %
        (self.type, self.help, self.default)
    )

  def __eq__(self, other):
    return super(optional, self).__eq__(other) and \
        isinstance(other, optional) and \
        other.default == self.default


class required(variable):
  """A required variable

  Used in a spec file to denote a variable that must be in a valid config file.
  If a required variable is not in a config file, validation will fail.

  >>> version = required(type=str, \
                         help="Version of this application")

  Parameters
  ----------
  type : function
      function to coerce argument to desired type
  help : str
      long description of this variable
  """
  def __init__(self, type=identity, help=None):
    super(required, self).__init__(type, help)

  def __str__(self):
    return (
        "required(type=%s, help=%s)" %
        (self.type, self.help)
    )

  def __eq__(self, other):
    return super(required, self).__eq__(other) and \
        isinstance(other, required)


def one_of(*options):
  """is this object one of these options?"""
  def type(obj):
    if not obj in options:
      raise ValueError("%s isn't one of %s" % (obj, options))
    return obj
  return type


def is_spec(obj):
  """Is this a valid specification for an object?"""
  return any(test(obj) for test, _ in VALIDATORS)


def validate_required(spec, config):
  if config is Missing:
    raise ValidationError("Missing required argument")
  if hasattr(spec.type, '__call__'):
    try:
      return spec.type(config)
    except ValueError:
      raise ValidationError('failed to convert "{}" with "{}"'.format(config, spec))
  else:
    # a spec, rather than a function, was used as the type.
    return _validate(spec.type, config)


def validate_optional(spec, config):
  if config is Missing:
    config = spec.default
  return _validate(required(type=spec.type), config)


def validate_dict(spec, config):
  if config is Missing:
    # if config is missing, replace it with an empty dict. if any of the spec's
    # contents are required, this will throw a ValidationError later; on the
    # other hand, if they're all optional, they'll be replaced sensibly.
    config = {}

  if not isinstance(config, dict):
    raise ValidationError('spec calls for type dict; found "{}" instead'.format(config))

  # for each key-value pair, replace with validated bit or what's already in
  # the config IF the value is in fact a spec definition
  result = {}
  for k, v in spec.items():
    if is_spec(v):
      result[k] = _validate(v, config.get(k, Missing))
  return result


def validate_list(spec, config):
  if len(spec) > 1:
    raise ValidationError('spec for list "{}" contains multiple definitions for its contents')
  elif len(spec) == 0:
    # no spec for contents, so just take it as everything is OK
    spec = [required(type=identity)]

  if config is Missing:
    # lists are always optional; if unspecified, an empty list is used
    config = []

  if not isinstance(config, list):
    raise ValidationError('spec calls for type list; found "{}" instead'.format(config))

  return [_validate(spec[0], c) for c in config]


def validate_tuple(spec, config):
  if config is Missing:
    # if config is missing, replace it a tuple of all missing values. if all of
    # this spec's fields are optional, then they'll be replaced reasonable;
    # otherwise, a ValidationError will be thrown, as expected.
    config = tuple([Missing] * len(spec))

  if not isinstance(config, tuple):
    if hasattr(config, '__iter__'):
      # coerce into a tuple. when using command line arguments
      # like '--a[1]', it's impossible to tell if a should be
      # a list or a tuple. this gives validation some leniency.
      config = tuple(config)
    else:
      raise ValidationError('spec calls for type tuple; found "{}" instead'.format(config))

  # extend config with Missing's if necessary
  if len(config) < len(spec):
    config = tuple(list(config) + [Missing] * (len(spec) - len(config)))

  # XXX what if spec and config have different lengths?
  if len(spec) != len(config):
    raise ValidationError('length of spec "{}" doesn\'t match config "{}"'.format(spec, config))

  return tuple( _validate(s, c) for s, c in zip(spec, config) )


VALIDATORS = [
    (lambda x: isinstance(x,required),validate_required),
    (lambda x: isinstance(x,optional),validate_optional),
    (lambda x: isinstance(x,    dict),    validate_dict),
    (lambda x: isinstance(x,    list),    validate_list),
    (lambda x: isinstance(x,   tuple),   validate_tuple),
  ]

def _validate(spec, config):
  for (test, func) in VALIDATORS:
    if test(spec):
      return func(spec, config)
  raise ValidationError('No validator for spec: "{}"'.format(spec))


def missing_required_keys(spec, config):
  result = []
  for k, v in spec.unroll().items():
    if isinstance(v, required) and k not in config:
      result.append(k)
  return result


def validate(spec, config):
  missing = missing_required_keys(spec, attrs.from_dict(config))
  if len(missing) > 0:
    text = "Missing required fields: " + ", ".join(missing)
    raise ValidationError("".join(text))

  return _validate(spec, config)



"""
Attribute dictionary
"""
import re

from .exceptions import ConfiguratiException
from .utils import recursive_apply, Missing


class adict(dict):
  """A dict you can access with dot notation"""

  def __getattr__(self, key):
    if key in self:
      return self[key]
    else:
      raise AttributeError("No attribute: " + str(key))

  def __setattr__(self, key, value):
    self[key] = value
    return self[key]

  def to_dict(self):
    f = lambda x: dict(x) if isinstance(x, attrs) else x
    return recursive_apply(self, value_func=f)

  @classmethod
  def from_dict(cls, d):
    f = lambda x: cls(x) if isinstance(x, dict) else x
    return recursive_apply(d, value_func=f)

  def unroll(self):
    unrolled = unroll(self)
    return { k[1:]:v for k, v in unrolled.items() }


class attrs(adict):
  """A dictionary/object designed for nested objects"""

  def __init__(self, *args, **kwargs):
    super(attrs, self).__init__()

    # get key, value pairs passed in
    if len(args) == 0:
      a = []
    elif len(args) == 1:
      a = args[0]
      if isinstance(a, dict):
        a = a.items()
    elif len(args) > 1:
      raise TypeError(
          "dict expected at most 1 arguments, got {}".format(len(args))
        )
    a += kwargs.items()

    # add them all, one by one
    for k, v in a:
      self[k] = v

  def __getitem__(self, key):
    if is_atomic(key):
      return super(attrs, self).__getitem__(key)
    else:
      return get(self, '.' + key)

  def __setitem__(self, key, value):
    if not valid_key(key):
      raise KeyError(
          "invalid attrs key: {}".format(key)
        )
    value = attrs.from_dict(value)
    if is_atomic(key):
      return super(attrs, self).__setitem__(key, value)
    else:
      return set(self, '.' + key, value, build=True)

  def __contains__(self, key):
    try:
      self[key]
      return True
    except:
      return False



def is_atomic(key):
  """return True is a key is "simple" in the sense of not referring to a nested
  object. e.g., 'a' is simple but 'a.b' is not"""
  return not ('[' in key or '.' in key)


def unroll(a):
  if isinstance(a, dict):
    result = {}
    for k, v in a.items():
      for k_, v_ in unroll(v).items():
        result["." + k + k_] = v_
    return result
  elif isinstance(a, tuple) or isinstance(a, list):
    result = {}
    for k, v, in enumerate(a):
      for k_, v_ in unroll(v).items():
        result["[{}]{}".format(k, k_)] = v_
    return result
  else:
    return {'': a}


def get(obj, key):
  """Retrieve a key from a nested object"""
  if len(key) == 0:
    return obj
  else:
    key, rest = next_key(key)
    if isinstance(obj, dict):
      return get(obj[key], rest)
    elif isinstance(obj, list) or isinstance(obj, tuple):
      if not isinstance(key, int):
        raise KeyError('Attempting to use non-integer index "{}" on list or tuple'.format(key))
      if key >= len(obj):
        raise KeyError('Attempting to access index {} but only {} available'.format(key, len(obj)))
      return get(obj[key], rest)
    else:
      raise KeyError('Still have additional key components "{}" but have already reached terminal node {}'.format(key, obj))


def set(obj, key, value, build=False):
  """Set a key to a value in a nested object"""
  if len(key) == 0:
    return value
  else:
    def set_dict(obj, key, value):
      obj[key] = set(obj.get(key, Missing), rest, value, build)
      return obj

    def set_list(obj, key, value):
      if not isinstance(key, int):
        raise KeyError('Attempting to use non-integer index "{}" on list or tuple'.format(key))
      if key >= len(obj):
        obj = obj + [Missing] * (key + 1 - len(obj))
      obj[key] = set(obj[key], rest, value, build)
      return obj

    def set_tuple(obj, key, value):
      return tuple(set_list(list(obj), key, value))

    # parse next atomic key
    key, rest = next_key(key)

    # make an object if `build` and object is missing
    if build and obj is Missing:
      if isinstance(key, basestring):
        obj = {}
      elif isinstance(key, int):
        obj = [Missing] * (key + 1)
      else:
        raise KeyError("Unrecognized key type: {}".format(str(type(key))))

    if isinstance(obj, dict):
      return set_dict(obj, key, value)
    elif isinstance(obj, list):
     return set_list(obj, key, value)
    elif isinstance(obj, tuple):
      return set_tuple(obj, key, value)
    else:
      raise KeyError('Still have additional key components "{}" but have already reached terminal node {}'.format(key, obj))


def next_key(s):
  if s.startswith('.'):
    # get "key_1" from ".key_1[0]" or ".key_1.key_2"
    match = re.search("""^[.]([^.[]*)""", s)
    key, rest = match.group(1), s[match.end():]
    if re.search("""^[a-zA-Z][a-zA-Z0-9-_]*$""", key) is None:
      raise KeyError('"{}" is not a attrs key'.format(key))
    return key.replace("-", "_"), rest
  elif s.startswith('['):
    # get "5" from "[5].key_1" or "[5][3]"
    match = re.search("""^\[([^[.]*)\]""", s)
    key, rest = match.group(1), s[match.end():]
    try:
      key = int(key)
    except ValueError:
      raise KeyError('"{}" is not a valid list/tuple key'.format(key))
    return key, rest
  else:
    raise KeyError('"{}" is not a valid attrs key'.format(s))


def valid_key(k):
  if not isinstance(k, basestring):
    return False
  else:
    k = '.' + k
    while len(k) > 0:
      try:
        _, k = next_key(k)
      except KeyError:
        return False
    return True
